import 'package:freezed_annotation/freezed_annotation.dart';
import 'package:mobx/mobx.dart';

part 'machine.freezed.dart';

typedef OnEnterState<StateValue, S extends StateValue, Event> = void Function(
    ActiveState<StateValue, S, Event> self);
typedef MachineSend<Event> = void Function(Event event);
typedef BuildMachine<StateValue, Event> = StateValue Function(
    MachineBuilder<StateValue, Event> builder);
typedef RemoveListener = void Function();

@freezed
class MachineStatus<StateValue, Event> with _$MachineStatus<StateValue, Event> {
  /// Indicates that the Machine is in idle. This means it has never been
  /// Stopped or Started before. The machine can not transition into idle.
  factory MachineStatus.idle() = MachineStatusIdle<StateValue, Event>;

  /// Machine is stopped at the moment. This means there is no active state.
  factory MachineStatus.stopped({
    required StateValue lastState,
  }) = MachineStatusStopped<StateValue, Event>;

  /// The machine is running and has an active state.
  factory MachineStatus.running({
    required StateValue state,
  }) = MachineStatusRunning<StateValue, Event>;
}

class MachineBuilder<StateValue, Event> {
  final List<StateNode<StateValue, Event>> _states = [];
  bool _autostart = false;

  /// Add a state to the machine. Also add a callback that is executed, once the
  /// machine reaches the state given.
  void onState<S extends StateValue>(OnEnterState<StateValue, S, Event> cb) {
    _states.add(StateNode<StateValue, Event>(
      enter: (state) => cb(state as ActiveState<StateValue, S, Event>),
      isState: (dynamic obj) => obj is S,
      createState: (m, v) => ActiveState<StateValue, S, Event>(
        state: v as S,
        machine: m,
      ),
    ));
  }

  /// Automatically start the machine during construct.
  void startAsRunning() => _autostart = true;
}

/// State Machine
class Machine<StateValue, Event> {
  Machine(BuildMachine<StateValue, Event> cb) {
    final builder = MachineBuilder<StateValue, Event>();
    _initialSV = cb(builder);
    _states = builder._states;
    _$state = Observable<MachineStatus<StateValue, Event>>(
      MachineStatus<StateValue, Event>.idle(),
      context: _ctx,
      name: '$runtimeType.state',
    );
    if (builder._autostart) {
      _transition(_initialSV);
    }
  }

  static final _ctx = ReactiveContext(config: ReactiveConfig.main);

  late final _$actionCtrl =
      ActionController(name: runtimeType.toString(), context: _ctx);
  late final List<StateNode<StateValue, Event>> _states;
  ActiveState<StateValue, StateValue, Event>? _activeState;
  late StateValue _initialSV;

  final List<ReactionDisposer> _disposer = [];
  bool get hasListener => _disposer.isNotEmpty;

  late Observable<MachineStatus<StateValue, Event>> _$state;
  MachineStatus<StateValue, Event> get state => _$state.value;

  bool _mounted = true;

  /// Whether [dispose] was called or not.
  bool get mounted => _mounted;

  /// Dispose the current machine. This will also clear the current active state,
  /// if any.
  void dispose() {
    _disposer
      ..forEach((disp) => disp())
      ..clear();
    _activeState?.dispose();
    _activeState = null;
    _states.clear();
    _mounted = false;
  }

  StateNode<StateValue, Event> _getNode(StateValue value) {
    return _states.firstWhere(
      (node) => node.isState(value),
      orElse: () =>
          throw Exception('StateValue "$value" not found in "$runtimeType" '),
    );
  }

  void _transition(StateValue value) {
    final _$actionInfo =
        _$actionCtrl.startAction(name: '$runtimeType.transition');
    try {
      final active = _activeState;
      _activeState = null;
      active?.dispose();

      final node = _getNode(value);
      final newActiveState = node.createState(this, value);
      _activeState = newActiveState;
      _$state.value = MachineStatus<StateValue, Event>.running(state: value);
      node.enter(newActiveState);
    } finally {
      _$actionCtrl.endAction(_$actionInfo);
    }
  }

  /// Check if it is possible to send and [event] to the current active state.
  bool canSend(Event event) => _$state.value.map(
      idle: (_) => false,
      stopped: (_) => false,
      running: (running) => null != _activeState?._candidate(event));

  /// Send an event to the current active state.
  /// This will throw if the machine is not running
  void send(Event event) => state.map(
      idle: (idle) => throw StateError(
          'Trying to send event "$event" to "$runtimeType" which is not possible because it is in Idle status'),
      stopped: (stopped) => throw StateError(
          'Trying to send event "$event" to "$runtimeType" which is not possible because it is stopped'),
      running: (running) {
        final candidate = _activeState?._candidate(event);
        if (null == candidate) return;
        candidate.handler(event);
      });

  /// Start the machine but only if it is not already running.
  /// Otherwise it will throw.
  void start({
    StateValue? stateValue,
  }) {
    _$state.value.maybeMap(
      running: (running) => throw StateError(
          'Trying to start "$runtimeType" which is already running'),
      orElse: () => _transition(stateValue ?? _initialSV),
    );
  }

  /// Stop the machine but only if it is  running.
  /// Otherwise it will throw.
  void stop() {
    final _$actionInfo = _$actionCtrl.startAction(name: '$runtimeType.stop');
    try {
      _activeState?.dispose();
      _activeState = null;

      _$state.value.map(
          idle: (idle) => throw StateError(
              'Trying to stop "$runtimeType" is not possible because it is in Idle status'),
          stopped: (stopped) => throw StateError(
              'Trying to stop "$runtimeType" is not possible because it is already stopped'),
          running: (running) {
            _$state.value = MachineStatus<StateValue, Event>.stopped(
              lastState: running.state,
            );
          });
    } finally {
      _$actionCtrl.endAction(_$actionInfo);
    }
  }

  RemoveListener addListener(
    void Function(MachineStatus<StateValue, Event> status) cb, {
    bool fireImmediately = true,
  }) {
    final disposer = reaction<MachineStatus<StateValue, Event>>(
      (_) => state,
      (s) {
        cb(s);
      },
      fireImmediately: fireImmediately,
      equals: (_, __) => false,
      context: _ctx,
      name: '$runtimeType.addListener',
      onError: (e, _) => throw e,
    );

    _disposer.add(disposer);

    return () {
      if (disposer.reaction.isDisposed) return;
      _disposer.removeWhere((element) => element == disposer);
      disposer();
    };
  }
}

@freezed
class StateNode<StateValue, Event> with _$StateNode<StateValue, Event> {
  factory StateNode({
    required void Function(ActiveState<StateValue, StateValue, Event> self)
        enter,
    required ActiveState<StateValue, StateValue, Event> Function(
      Machine<StateValue, Event> machine,
      StateValue value,
    )
        createState,
    required bool Function(dynamic obj) isState,
  }) = _StateNode<StateValue, Event>;
}

@freezed
class EventWrapper<Event> with _$EventWrapper<Event> {
  factory EventWrapper({
    required void Function(Event event) handler,
    required bool Function(Event obj) isCandidate,
  }) = _EventWrapper<Event>;
}

class ActiveState<StateValue, S, Event> {
  ActiveState({
    required Machine<StateValue, Event> machine,
    required this.state,
  }) : _machine = machine;

  final Machine<StateValue, Event> _machine;
  final List<void Function()> _onExit = [];
  final List<EventWrapper<Event>> _onEvent = [];

  /// The state when entered
  final S state;

  /// Check if the current state is still active.
  bool get isActive => _machine._activeState == this;

  void dispose() {
    _onEvent.clear();
    _onExit
      ..forEach((exitCb) => exitCb())
      ..clear();
  }

  EventWrapper<Event>? _candidate(Event event) {
    return _onEvent.firstWhereOrNull((element) => element.isCandidate(event));
  }

  /// Transition to another state. This can only be called once in the current state.
  /// Subsequent calls are ignored.
  void transition(StateValue value) {
    if (!isActive) {
      throw StateError(
          'Cannot transition because $StateValue is no longer active');
    }

    _machine._transition(value);
  }

  /// Execute the callback given when leaving the state.
  /// Calls are immediately executed after state has been left.
  void onExit(void Function() cb) {
    if (!isActive) {
      throw StateError(
          'Cannot add an exit callback because $StateValue is no longer active');
    }

    _onExit.add(cb);
  }

  /// Attach an event to the current active state. This way one may communicate with the
  /// StateMachine from outside.
  /// Usually this is where a [transition] should happen.
  void onEvent<E extends Event>(void Function(E event) cb) {
    if (!isActive) {
      throw StateError(
          'Cannot add an event callback because $StateValue is no longer active');
    }

    _onEvent.add(EventWrapper<Event>(
      handler: (event) => cb(event as E),
      isCandidate: (dynamic obj) => obj is E,
    ));
  }

  /// Nest a machine. The nested machine will be automatically started by either
  /// its default initial state value or by [startWith].
  /// The nested machine will automatically be stopped when the state is left.
  /// It is possible to listen to state changes of the nested machine using [onState].
  /// It is fine to transition to another state in [onState].
  /// [onUncaughtError] will be called when there is an uncaught exception in [onState].
  ///
  /// Calling this method while the state is no longer active will do nothing.
  /// It is an error to nest a already running macine.
  void nest<ChildStateValue>(Machine<ChildStateValue, dynamic> machine,
      {ChildStateValue? startWith,
      void Function(ChildStateValue stateValue)? onState}) {
    if (!isActive) return;

    if (null != onState) {
      final disposer = reaction<MachineStatus<ChildStateValue, dynamic>>(
        (_) => machine.state,
        (status) {
          status.maybeMap(
            orElse: () {},
            running: (running) => onState(running.state),
          );
        },
        context: Machine._ctx,
        fireImmediately: true,
        equals: (_, __) => false,
        name: '$runtimeType.nested.${machine.runtimeType}}',
        onError: (obj, r) => throw obj,
      );
      onExit(() => disposer());
    }

    machine.state.map(
      idle: (idle) => machine.start(stateValue: startWith),
      stopped: (stopped) => machine.start(stateValue: startWith),
      running: (running) => throw StateError('machine must not be running'),
    );

    onExit(() => machine.stop());
  }
}
