import 'dart:async';

import 'package:freezed_annotation/freezed_annotation.dart';
import 'package:mockito/annotations.dart';
import 'package:mockito/mockito.dart';
import 'package:test/test.dart';
import 'package:endlich/src/machine.dart';

import 'endlich_test.mocks.dart';

part 'endlich_test.freezed.dart';

@GenerateMocks(
    [EventListener, OnExitCb, OnActiveStateCb, OnStateCb, AddListenerCb])
void main() {
  Machine<State, Event> fooMachine({
    State? initial,
    bool? autostart,
    void Function(ActiveState<State, _Foo, Event> self)? onFoo,
    void Function(ActiveState<State, _Bar, Event> self)? onBar,
    void Function(ActiveState<State, _Baz, Event> self)? onBaz,
  }) {
    return Machine<State, Event>((b) {
      b.onState<_Foo>((self) {
        if (null != onFoo) onFoo(self);
      });
      b.onState<_Bar>((self) {
        if (null != onBar) onBar(self);
      });
      b.onState<_Baz>((self) {
        if (null != onBaz) onBaz(self);
      });
      if (true == autostart) {
        b.startAsRunning();
      }
      return initial ?? const State.foo();
    });
  }

  test('create machine', () {
    final machine = fooMachine();
    expect(machine.state, isA<MachineStatusIdle<State, Event>>());
  });

  test('create machine auto started', () {
    final enterFoo = MockOnActiveStateCb();
    final machine = fooMachine(autostart: true, onFoo: enterFoo);

    expect(
        machine.state,
        isA<MachineStatusRunning<State, Event>>()
            .having((s) => s.state, 'state', const State.foo()));
    verify(enterFoo(any)).called(1);
  });

  test('change status', () {
    final machine = fooMachine();

    machine.start();
    expect(
        machine.state,
        isA<MachineStatusRunning<State, Event>>()
            .having((s) => s.state, 'state', const State.foo()));

    machine.stop();
    expect(
        machine.state,
        isA<MachineStatusStopped<State, Event>>()
            .having((s) => s.lastState, 'lastState', const State.foo()));

    machine.start(stateValue: const State.baz());
    expect(
        machine.state,
        isA<MachineStatusRunning<State, Event>>()
            .having((s) => s.state, 'state', const State.baz()));

    machine.stop();
    expect(
        machine.state,
        isA<MachineStatusStopped<State, Event>>()
            .having((s) => s.lastState, 'lastState', const State.baz()));
  });

  test('.canSend()', () {
    final machine = fooMachine(onFoo: (self) {
      self.onEvent<_Next>((event) {});
    }, onBar: (self) {
      self.onEvent<_ToBaz>((event) {});
    });

    expect(machine.canSend(const Event.next()), isFalse);
    machine.start();
    expect(machine.canSend(const Event.next()), isTrue);
    expect(machine.canSend(const Event.toFoo()), isFalse);

    machine.stop();
    expect(machine.canSend(const Event.next()), isFalse);

    machine.start(stateValue: const State.bar());
    expect(machine.canSend(const Event.next()), isFalse);
    expect(machine.canSend(const Event.toBaz()), isTrue);
  });

  test('enter a new state', () {
    final enterFoo = MockOnActiveStateCb<_Foo>();
    final enterBar = MockOnActiveStateCb<_Bar>();

    final machine = fooMachine(
      onFoo: (self) {
        enterFoo(self);
        self.onEvent<_Next>((event) {
          self.transition(const State.bar());
        });
      },
      onBar: (self) {
        enterBar(self);
      },
    );
    machine.start();
    machine.send(const Event.next());
    verifyInOrder([
      enterFoo(argThat(isA<ActiveState<State, _Foo, Event>>()
          .having((s) => s.state, 'state when entered', const State.foo()))),
      enterBar(argThat(isA<ActiveState<State, _Bar, Event>>()
          .having((s) => s.state, 'state when entered', const State.bar())))
    ]);
  });

  group('ActiveState.isActive', () {
    test('after leaving state', () {
      late ActiveState<State, _Foo, Event> foo;
      final machine = fooMachine(onFoo: (self) {
        foo = self;
        self.onEvent<_Next>((event) {
          self.transition(const State.bar());
        });
      })
        ..start();
      expect(foo.isActive, isTrue);
      machine.send(const Event.next());
      expect(foo.isActive, isFalse);
    });

    test('after machine status became stopped', () {
      late ActiveState<State, _Foo, Event> foo;
      fooMachine(onFoo: (self) {
        foo = self;
      })
        ..start()
        ..stop();

      expect(foo.isActive, isFalse);
    });

    test('after machine has been disposed', () {
      late ActiveState<State, _Foo, Event> foo;
      fooMachine(onFoo: (self) {
        foo = self;
      })
        ..start()
        ..dispose();

      expect(foo.isActive, isFalse);
    });
  });

  group('ActiveState..onExit()', () {
    test('callback will be executed when leaving state', () {
      final leavingFoo1 = MockOnExitCb();
      final leavingFoo2 = MockOnExitCb();
      final leavingBar1 = MockOnExitCb();
      final machine = fooMachine(onFoo: (self) {
        self.onEvent<_Next>((event) {
          self.transition(const State.bar());
        });
        self.onExit(leavingFoo1);
        self.onExit(leavingFoo2);
      }, onBar: (self) {
        self.onExit(leavingBar1);
      });
      machine
        ..start()
        ..send(const Event.next())
        ..stop()
        ..start()
        ..dispose();

      verifyInOrder([
        leavingFoo1(),
        leavingFoo2(),
        leavingBar1(),
        leavingFoo1(),
        leavingFoo2(),
      ]);
    });

    test('does not allow to execute helper functions in body', () {
      dynamic transitionErr;
      dynamic onEventErr;
      dynamic onExitErr;
      fooMachine(onFoo: (self) {
        self.onEvent<_Next>((event) {
          self.transition(const State.bar());
        });
        self.onExit(() {
          try {
            self.transition(const State.bar());
          } catch (e) {
            transitionErr = e;
          }

          try {
            self.onEvent<_Next>((event) {});
          } catch (e) {
            onEventErr = e;
          }

          try {
            self.onExit(() {});
          } catch (e) {
            onExitErr = e;
          }
        });
      })
        ..start()
        ..send(const Event.next());

      expect(transitionErr, isStateError);
      expect(onEventErr, isStateError);
      expect(onExitErr, isStateError);
    });
  });

  test('transition on event', () {
    final machine = fooMachine(onFoo: (self) {
      self.onEvent<_Next>((event) {
        self.transition(const State.bar());
      });
    });

    machine.start();
    machine.send(const Event.next());
    expect(
        machine.state,
        isA<MachineStatusRunning<State, Event>>()
            .having((s) => s.state, 'state', const State.bar()));
  });

  test('transition immediately', () {
    final machine = fooMachine(onFoo: (self) {
      self.transition(const State.bar());
    }, onBar: (self) {
      self.transition(const State.baz());
    });

    machine.start();
    expect(
        machine.state,
        isA<MachineStatusRunning<State, Event>>()
            .having((s) => s.state, 'state', const State.baz()));
  });

  test('transition in async', () async {
    final c = Completer<void>();
    final machine = fooMachine(onFoo: (self) {
      Future(() {
        self.transition(const State.bar());
        c.complete();
      });
    });

    machine.start();
    await c.future;
    expect(
        machine.state,
        isA<MachineStatusRunning<State, Event>>()
            .having((s) => s.state, 'state', const State.bar()));
  });

  test(
      'throw error on ActiveState helper functions, after leaving/disposing the state',
      () async {
    dynamic transitionErr;
    dynamic onEventErr;
    dynamic onExitErr;
    final machine = fooMachine(onFoo: (self) {
      self.transition(const State.bar());
      try {
        self.transition(const State.baz());
      } catch (e) {
        transitionErr = e;
      }

      try {
        self.onEvent<_Next>((event) {});
      } catch (e) {
        onEventErr = e;
      }

      try {
        self.onExit(() {});
      } catch (e) {
        onExitErr = e;
      }
    });

    machine.start();
    expect(transitionErr, isStateError);
    expect(onEventErr, isStateError);
    expect(onExitErr, isStateError);
  });

  test('register event on active state', () {
    final fooNext = MockEventListener<_Next>();
    final fooToBaz = MockEventListener<_ToBaz>();
    final machine = fooMachine(onFoo: (self) {
      self.onEvent<_Next>(fooNext);
      self.onEvent<_ToBaz>(fooToBaz);
      self.onEvent<_ToBar>((event) {
        self.transition(const State.bar());
      });
    });
    machine.start();

    machine.send(const Event.next());
    machine.send(const Event.toBaz());
    machine.send(const Event.next());

    expect(machine.canSend(const Event.next()), isTrue);
    expect(machine.canSend(const Event.toBaz()), isTrue);
    expect(machine.canSend(const Event.toBar()), isTrue);
    expect(machine.canSend(const Event.toFoo()), isFalse);

    machine.send(const Event.toBar());

    expect(machine.canSend(const Event.next()), isFalse);
    expect(machine.canSend(const Event.toBaz()), isFalse);
    expect(machine.canSend(const Event.toBar()), isFalse);
    expect(machine.canSend(const Event.toFoo()), isFalse);

    machine.send(const Event.next());
    machine.send(const Event.next());

    verifyInOrder([
      fooNext(argThat(isA<_Next>())),
      fooToBaz(argThat(isA<_ToBaz>())),
      fooNext(argThat(isA<_Next>())),
    ]);
  });

  group('mounted', () {
    test('is no longer mounted after dispose', () {
      final machine = fooMachine();
      expect(machine.mounted, isTrue);
      machine.dispose();
      expect(machine.mounted, isFalse);
    });
  });

  group('listen to machine', () {
    test('can be unsubscribed either manually or during machine dispose', () {
      final listener = MockAddListenerCb();
      final listener2 = MockAddListenerCb();

      final machine = fooMachine(onFoo: (self) {
        self.onEvent<_Next>((event) {
          self.transition(const State.bar());
        });
      });
      expect(machine.hasListener, isFalse);

      final unsub = machine.addListener(listener);
      machine.addListener(listener2);

      expect(machine.hasListener, isTrue);

      unsub();
      expect(machine.hasListener, isTrue);

      machine.dispose();

      expect(machine.hasListener, isFalse);
      expect(() => unsub(), isNot(throwsException));
    });

    test('can fire immediately', () {
      final listener = MockAddListenerCb();
      final listener2 = MockAddListenerCb();
      fooMachine().addListener(listener);
      fooMachine().addListener(listener2, fireImmediately: false);

      verifyInOrder([
        listener(argThat(isA<MachineStatusIdle<State, Event>>())),
      ]);
      verifyZeroInteractions(listener2);
    });

    test('on every status change', () {
      final listener = MockAddListenerCb();
      fooMachine(onFoo: (self) {
        self.onEvent<_Next>((event) {
          self.transition(const State.bar());
        });
      })
        ..addListener(listener)
        ..start()
        ..send(const Event.next())
        ..stop();

      verifyInOrder([
        listener(argThat(isA<MachineStatusIdle<State, Event>>())),
        listener(argThat(isA<MachineStatusRunning<State, Event>>()
            .having((s) => s.state, 'state', const State.foo()))),
        listener(argThat(isA<MachineStatusRunning<State, Event>>()
            .having((s) => s.state, 'state', const State.bar()))),
        listener(argThat(isA<MachineStatusStopped<State, Event>>())),
      ]);
    });
  });

  group('nested', () {
    test('child will start/stop', () {
      final child = fooMachine(onFoo: (self) {});
      final parent = fooMachine(onFoo: (self) {
        self.onEvent<_Next>((event) => self.transition(const State.bar()));
        self.nest<State>(child);
      })
        ..start();

      expect(
          child.state,
          isA<MachineStatusRunning<State, Event>>()
              .having((s) => s.state, 'state', const State.foo()));

      parent.stop();
      expect(
          child.state,
          isA<MachineStatusStopped<State, Event>>()
              .having((s) => s.lastState, 'state', const State.foo()));

      parent.start();
      expect(
          child.state,
          isA<MachineStatusRunning<State, Event>>()
              .having((s) => s.state, 'state', const State.foo()));

      parent.send(const Event.next());
      expect(
          child.state,
          isA<MachineStatusStopped<State, Event>>()
              .having((s) => s.lastState, 'state', const State.foo()));

      parent
        ..stop()
        ..start(stateValue: const State.foo());
      expect(
          child.state,
          isA<MachineStatusRunning<State, Event>>()
              .having((s) => s.state, 'state', const State.foo()));

      parent.dispose();
      isA<MachineStatusStopped<State, Event>>()
          .having((s) => s.lastState, 'state', const State.foo());
    });

    test('can start child with other state', () {
      final child = fooMachine(onFoo: (self) {});
      fooMachine(onFoo: (self) {
        self.onEvent<_Next>((event) => self.transition(const State.bar()));
        self.nest<State>(child, startWith: const State.bar());
      }).start();

      expect(
          child.state,
          isA<MachineStatusRunning<State, Event>>()
              .having((s) => s.state, 'state', const State.bar()));
    });

    test('can be listened to by parent', () {
      final deeperChildCb = MockOnStateCb();
      final childCb = MockOnStateCb();
      final parentCb = MockOnStateCb();
      final deeperChild = fooMachine(initial: const State.foo());
      final child = fooMachine(
          initial: const State.bar(),
          onBar: (self) {
            self.nest<State>(deeperChild, onState: deeperChildCb);
          });
      final parent = fooMachine(
          initial: const State.baz(),
          onBaz: (self) {
            self.nest<State>(child, onState: childCb);
          });

      parent.addListener((status) {
        status.maybeMap(
            orElse: () {}, running: (runnign) => parentCb(runnign.state));
      });

      parent.start();

      verifyInOrder([
        parentCb(argThat(isA<_Baz>())),
        childCb(argThat(isA<_Bar>())),
        deeperChildCb(argThat(isA<_Foo>())),
      ]);
    });

    test('transition in onState', () {
      final child = fooMachine(initial: const State.bar());
      final parent = fooMachine(onFoo: (self) {
        self.nest<State>(child, onState: (s) {
          self.transition(const State.bar());
        });
      })
        ..start();

      expect(
          parent.state,
          isA<MachineStatusRunning<State, Event>>()
              .having((s) => s.state, 'state', const State.bar()));
    });

    test('throws if nested machine is already running', () {
      final child = fooMachine();
      final parent = fooMachine(onFoo: (self) {
        self.nest<State>(child);
      });

      child.start();
      expect(() => parent.start(), throwsStateError);
    });
  });
}

@freezed
class State with _$State {
  const factory State.foo() = _Foo;
  const factory State.bar() = _Bar;
  const factory State.baz() = _Baz;
}

@freezed
class Event with _$Event {
  const factory Event.next() = _Next;
  const factory Event.toFoo() = _ToFoo;
  const factory Event.toBar() = _ToBar;
  const factory Event.toBaz() = _ToBaz;
}

class EventListener<Event> extends Mock {
  void call(Event event);
}

class OnExitCb extends Mock {
  void call();
}

class OnActiveStateCb<S extends State> extends Mock {
  void call(ActiveState<State, S, Event> self);
}

class OnStateCb extends Mock {
  void call(State state);
}

class AddListenerCb extends Mock {
  void call(MachineStatus<State, Event> status);
}
